from __future__ import annotations
"""Provides the module functions
"""
import json
from typing import Any, Optional
import urllib.parse
from concurrent.futures import as_completed, ThreadPoolExecutor
from sys import exit as sysexit
import time
import os
import shutil
import requests
from PIL import Image
from emoji import demojize
import pylottie

opj = os.path.join


def assureDirExists(directory: str, root: str) -> str:
	"""make the dir if not exists

	Args:
		dir (str): the directory name
		root (str): the path of the root directory

	Returns:
		str: the full path
	"""
	fullPath = opj(root, directory)
	if os.path.isdir(fullPath):
		pass
	else:
		os.mkdir(fullPath)

	return fullPath


class Sticker:
	"""Sticker instance attributes
	"""
	def __init__(self, name: str = "None", link: str = "None", emoji: str = "😀",
	animated: bool = False):
		self.name = name
		self.link = link
		self.emoji = emoji
		self.animated = animated

	def __repr__(self):
		return "<F:{}>".format(self.name)

	def emojiName(self) -> str:
		""" get the emoji as a string """
		return demojize(self.emoji)[1:-1]


class StickerDownloader:
	"""The StickerDownloader sets up the api and makes requests
	"""
	def __init__(self, token, session=None, multithreading=4):
		self.threads = multithreading
		self.token = token
		self.cwd = assureDirExists("downloads", root=os.getcwd())
		if session is None:
			self.session = requests.Session()
		else:
			self.session = session
		self.api = "https://api.telegram.org/bot{}/".format(self.token)
		verify = self.doAPIReq("getMe", {})
		if verify is not None and verify["ok"]:
			pass
		else:
			print("Invalid token.")
			sysexit()

	def doAPIReq(self, fstring: str, params: dict[Any, Any]) -> Optional[dict[Any,
	Any]]:
		"""Use the telegram api

		Args:
			fstring (str): function to execute
			params (dict[Any, Any]): function parameters

		Raises:
			RuntimeError: In the event of a failure

		Returns:
			Optional[dict[Any, Any]]: api response
		"""
		try:
			urlParams = "?" + urllib.parse.urlencode(params)
			res = self.session.get("{}{}{}".format(self.api, fstring, urlParams))
			if res.status_code != 200:
				raise RuntimeError
			res = json.loads(res.content.decode("utf-8"))
			if not res["ok"]:
				raise RuntimeError(res["description"])
			return res
		except RuntimeError as exception:
			print("API method {} failed. Error: \"{}\"".format(fstring, exception))
			return None

	def getSticker(self, fileData: dict[Any, Any]) -> Sticker:
		"""Get sticker info from the server

		Args:
			fileData (dict[Any, Any]): sticker id

		Returns:
			Sticker: Sticker instance
		"""
		info = self.doAPIReq("getFile", {"file_id": fileData["file_id"]})
		if info is not None:
			file = Sticker(
			name=info["result"]["file_path"].split("/")[-1],
			link="https://api.telegram.org/file/bot{}/{}"
			.format(self.token, info["result"]["file_path"]), emoji=fileData["emoji"],
			animated=fileData["is_animated"])
			return file
		return Sticker()

	def getStickerSet(self, name: str) -> Optional[dict[Any, Any]]:
		"""	Get a list of File objects.

		Args:
			name (str): name of the sticker set

		Returns:
			dict[Any, Any]: dictionary containing sticker data
		"""
		params = {"name": name}
		res = self.doAPIReq("getStickerSet", params)
		if res is None:
			return None
		stickers = res["result"]["stickers"]
		files = []

		print("Starting to scrape \"{}\" ..".format(name))
		start = time.time()
		with ThreadPoolExecutor(max_workers=self.threads) as executor:
			futures = [executor.submit(self.getSticker, i) for i in stickers]
			for i in as_completed(futures):
				files.append(i.result())
		end = time.time()
		print("Time taken to scrape {} stickers - {:.3f}s"
		.format(len(files), end - start))
		print()

		stickerSet = {
		"name": res["result"]["name"].lower(), "title": res["result"]["title"],
		"files": files}
		return stickerSet

	def downloadSticker(self, name: str, link: str, path: str) -> str:
		"""Download a sticker from the server

		Args:
			name (str): the name of the file
			link (str): the url to the file on the server
			path (str): the path to write to

		Returns:
			str: the filepath the file was written to
		"""
		filePath = opj(path, name)
		with open(filePath, "wb") as file:
			res = self.session.get(link)
			file.write(res.content)
		return filePath

	def downloadStickerSet(self, stickerSet: dict[Any, Any]):
		"""
		Download sticker set.
		"""
		swd = assureDirExists(stickerSet["name"], root=self.cwd)
		webpDir = assureDirExists("webp", root=swd)
		tgsDir = assureDirExists("tgs", root=swd)
		downloads = []

		print("Starting download of \"{}\" into {}".format(stickerSet["name"], swd))
		start = time.time()
		with ThreadPoolExecutor(max_workers=self.threads) as executor:
			futures = [executor.submit(self.downloadSticker,
			name="{}+{}+{}.{}".format(sticker.name.split("_")[-1].split(".")[0],
			sticker.emojiName(), sticker.emoji,
			("tgs" if sticker.animated else "webp")),
			link=sticker.link,
			path=tgsDir if sticker.animated else webpDir)
			for sticker in stickerSet["files"]] # yapf: disable
			for i in as_completed(futures):
				downloads.append(i.result())

		end = time.time()
		print("Time taken to download {} stickers - {:.3f}s"
		.format(len(downloads), end - start))
		print()
		return downloads

	def convertStatic(self, inputFile: str):
		"""Convert the webp file to png

		Args:
			inputFile (str): path to input file

		Returns:
			None
		"""
		img = Image.open(inputFile)
		img.save(inputFile.replace("webp", "png"))
		try:
			img.save(inputFile.replace("webp", "gif"), transparency=0,
			save_all=True, optimize=False) # yapf: disable
		except ValueError:
			print("Failed to save {} as gif".format(inputFile))

	def convertDir(self, name: str, quality: int = 1):
		"""	Convert the webp images into png images

		Args:
			name (str): name of the directory to convert
			quality (int): quality of animated images. Default=1
		"""
		# yapf: disable
		# Make directories
		swd = assureDirExists(name, root=self.cwd)
		webpDir = assureDirExists("webp", root=swd)
		tgsDir = assureDirExists("tgs", root=swd)
		assureDirExists("png", root=swd)
		assureDirExists("gif", root=swd)
		# List of animated stickers
		animatedStickers = [opj(tgsDir, i) for i in os.listdir(tgsDir)
		if i.endswith(".tgs")]
		# List of static stickers
		staticStickers = [opj(webpDir, i) for i in os.listdir(webpDir)]
		# Convert Stickers
		print("Converting stickers \"{}\"...".format(name))
		converted = 0
		start = time.time()
		# 	Static
		with ThreadPoolExecutor(max_workers=self.threads) as executor:
			futures = [executor.submit(self.convertStatic, inputFile)
			for inputFile in staticStickers]
			for _i in as_completed(futures):
				converted += 1
		# 	Animated
		if len(animatedStickers) > 0:
			imageDataList = pylottie.convertLotties2PIL(animatedStickers, quality)
			for index, imageData in enumerate(imageDataList):
				images = imageData[0]
				duration = imageData[1]
				converted += 1
				images[0].save(animatedStickers[index].replace("tgs", "gif"), save_all=True, append_images=images[1:],
				duration=duration*1000/len(images), loop=0, transparency=0, disposal=2)
				images[0].save(animatedStickers[index].replace("tgs", "webp"), save_all=True, append_images=images[1:],
				duration=int(duration*1000/len(images)), loop=0)
			shutil.rmtree("temp", ignore_errors=True)
		end = time.time()
		print("Time taken to convert {} stickers - {:.3f}s".format(converted, end - start))
		print()
		# yapf: enable



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['tstickers']

package_data = \
{'': ['*']}

install_requires = \
['Pillow>=8.1.0,<10',
 'emoji>=0.6.0,<2',
 'pylottie>=2021.2.1,<2023',
 'requests>=2.25.1,<4']

entry_points = \
{'console_scripts': ['tstickers = tstickers:cli']}

setup_kwargs = {
    'name': 'tstickers',
    'version': '2021.2',
    'description': 'Download sticker packs from Telegram',
    'long_description': '[![GitHub top language](https://img.shields.io/github/languages/top/FHPythonUtils/TStickers.svg?style=for-the-badge)](../../)\n[![Codacy grade](https://img.shields.io/codacy/grade/[projid].svg?style=for-the-badge)](https://www.codacy.com/gh/FHPythonUtils/TStickers)\n[![Repository size](https://img.shields.io/github/repo-size/FHPythonUtils/TStickers.svg?style=for-the-badge)](../../)\n[![Issues](https://img.shields.io/github/issues/FHPythonUtils/TStickers.svg?style=for-the-badge)](../../issues)\n[![License](https://img.shields.io/github/license/FHPythonUtils/TStickers.svg?style=for-the-badge)](/LICENSE.md)\n[![Commit activity](https://img.shields.io/github/commit-activity/m/FHPythonUtils/TStickers.svg?style=for-the-badge)](../../commits/master)\n[![Last commit](https://img.shields.io/github/last-commit/FHPythonUtils/TStickers.svg?style=for-the-badge)](../../commits/master)\n[![PyPI Downloads](https://img.shields.io/pypi/dm/tstickers.svg?style=for-the-badge)](https://pypi.org/project/tstickers/)\n[![PyPI Version](https://img.shields.io/pypi/v/tstickers.svg?style=for-the-badge)](https://pypi.org/project/tstickers/)\n\n<!-- omit in TOC -->\n# TStickers - Telegram Sticker Downloader\n\n<img src="readme-assets/icons/name.png" alt="Project Icon" width="750">\n\nDownload sticker packs from Telegram\n\nNOTE: You need a telegram bot token to make use of the script. Generate a bot\ntoken and paste in a file called \'env\'. Send a message to @BotFather to get started.\n\n- [Docs](#docs)\n- [Using](#using)\n- [Formats](#formats)\n- [Install With PIP](#install-with-pip)\n- [Language information](#language-information)\n\t- [Built for](#built-for)\n- [Install Python on Windows](#install-python-on-windows)\n\t- [Chocolatey](#chocolatey)\n\t- [Download](#download)\n- [Install Python on Linux](#install-python-on-linux)\n\t- [Apt](#apt)\n- [How to run](#how-to-run)\n\t- [With VSCode](#with-vscode)\n\t- [From the Terminal](#from-the-terminal)\n- [Download](#download-1)\n\t- [Clone](#clone)\n\t\t- [Using The Command Line](#using-the-command-line)\n\t\t- [Using GitHub Desktop](#using-github-desktop)\n\t- [Download Zip File](#download-zip-file)\n- [Community Files](#community-files)\n\t- [Licence](#licence)\n\t- [Changelog](#changelog)\n\t- [Code of Conduct](#code-of-conduct)\n\t- [Contributing](#contributing)\n\t- [Security](#security)\n\n## Docs\nSee the [Docs](/DOCS/README.md) for more information.\n\n\n## Using\n\n- Create a file called \'env\' (not env.txt) and paste your token\n- Get the URL of the telegram sticker pack\n- Run the program `python -m tstickers`\n- Enter the URL of the sticker pack\n- Get the output in the `downloads` folder.\n\n\n## Formats\n\n|Format|Static|Animated|\n|------|------|--------|\n|.gif  |✔     |✔      |\n|.png  |✔     |❌     |\n|.tgs  |❌    |✔      |\n|.webp |✔     |✔      |\n\n\nNote that static images can fail to save as .gif occasionally in testing\n\n## Install With PIP\n\n```python\npip install tstickers\n```\n\nHead to https://pypi.org/project/tstickers/ for more info\n\n\n## Language information\n### Built for\nThis program has been written for Python 3 and has been tested with\nPython version 3.9.0 <https://www.python.org/downloads/release/python-380/>.\n\n## Install Python on Windows\n### Chocolatey\n```powershell\nchoco install python\n```\n### Download\nTo install Python, go to <https://www.python.org/> and download the latest\nversion.\n\n## Install Python on Linux\n### Apt\n```bash\nsudo apt install python3.9\n```\n\n## How to run\n### With VSCode\n1. Open the .py file in vscode\n2. Ensure a python 3.9 interpreter is selected (Ctrl+Shift+P > Python:Select\nInterpreter > Python 3.9)\n3. Run by pressing Ctrl+F5 (if you are prompted to install any modules, accept)\n### From the Terminal\n```bash\n./[file].py\n```\n\n\n## Download\n### Clone\n#### Using The Command Line\n1. Press the Clone or download button in the top right\n2. Copy the URL (link)\n3. Open the command line and change directory to where you wish to\nclone to\n4. Type \'git clone\' followed by URL in step 2\n```bash\n$ git clone https://github.com/FHPythonUtils/TStickers\n```\n\nMore information can be found at\n<https://help.github.com/en/articles/cloning-a-repository>\n\n#### Using GitHub Desktop\n1. Press the Clone or download button in the top right\n2. Click open in desktop\n3. Choose the path for where you want and click Clone\n\nMore information can be found at\n<https://help.github.com/en/desktop/contributing-to-projects/cloning-a-repository-from-github-to-github-desktop>\n\n### Download Zip File\n\n1. Download this GitHub repository\n2. Extract the zip archive\n3. Copy/ move to the desired location\n\n## Community Files\n### Licence\nMIT License\n(See the [LICENSE](/LICENSE.md) for more information.)\n\n### Changelog\nSee the [Changelog](/CHANGELOG.md) for more information.\n\n### Code of Conduct\nIn the interest of fostering an open and welcoming environment, we\nas contributors and maintainers pledge to make participation in our\nproject and our community a harassment-free experience for everyone.\nPlease see the\n[Code of Conduct](https://github.com/FHPythonUtils/.github/blob/master/CODE_OF_CONDUCT.md) for more information.\n\n### Contributing\nContributions are welcome, please see the [Contributing Guidelines](https://github.com/FHPythonUtils/.github/blob/master/CONTRIBUTING.md) for more information.\n\n### Security\nThank you for improving the security of the project, please see the [Security Policy](https://github.com/FHPythonUtils/.github/blob/master/SECURITY.md) for more information.\n',
    'author': 'FredHappyface',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/FHPythonUtils/TStickers',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



"""Download sticker packs from Telegram
"""
import os
import argparse
from sys import exit as sysexit
from tstickers.downloader import StickerDownloader


def cli():
	""" cli entry point """
	parser = argparse.ArgumentParser("Welcome to TSticker, providing all of your sticker needs")
	parser.add_argument("-t", "--token", help="Pass in a bot token inline")
	parser.add_argument("-p", "--pack", help="Pass in a pack url inline", action="append")
	parser.add_argument("-q", "--quality", help="Set animation quality. default=1", type=int, default=1)
	args = parser.parse_args()
	# Get the token
	token = args.token
	if args.token is None:
		try:
			token = open(os.getcwd() + "/env", encoding="utf-8").readline().strip()
		except FileNotFoundError:
			try:
				token = open(os.getcwd() + "/env.txt", encoding="utf-8").readline().strip()
			except FileNotFoundError:
				print("!! Generate a bot token and paste in a file called \"env\". Send a "+
				"message to @BotFather to get started")
				sysexit(1)
	# Get the packs
	names = args.pack
	if names is None:
		names = []
		while True:
			name = input("Enter sticker_set url (leave blank to stop): ").strip()
			if name == "":
				break
			names.append(name)
	names = [name.split("/")[-1] for name in names]
	downloader = StickerDownloader(token)
	for sset in names:
		print("=" * 60)
		stickerSet = downloader.getStickerSet(sset)
		if stickerSet is None:
			continue
		print("-" * 60)
		_ = downloader.downloadStickerSet(stickerSet)
		print("-" * 60)
		downloader.convertDir(sset, args.quality)

if __name__ == "__main__":
	cli()
